import pkg_resources
from Corpus.Sentence import Sentence
from Dictionary.Word import Word
from MorphologicalAnalysis.FsmMorphologicalAnalyzer import FsmMorphologicalAnalyzer
from MorphologicalAnalysis.FsmParseList import FsmParseList
from NGram.NGram import NGram

from SpellChecker.Candidate import Candidate
from SpellChecker.NGramSpellChecker import NGramSpellChecker
from SpellChecker.Operator import Operator
from SpellChecker.SpellCheckerParameter import SpellCheckerParameter


class ContextBasedSpellChecker(NGramSpellChecker):
    __context_list: dict

    def __init__(self,
                 fsm: FsmMorphologicalAnalyzer,
                 nGram: NGram,
                 parameter: SpellCheckerParameter):
        """
        A constructor of {@link ContextBasedSpellChecker} class which takes a {@link FsmMorphologicalAnalyzer}, an {@link NGram}
        and a {@link SpellCheckerParameter} as inputs. Then, calls its super class {@link NGramSpellChecker} with given inputs.
        :param fsm: {@link FsmMorphologicalAnalyzer} type input.
        :param nGram: {@link NGram} type input.
        :param parameter: {@link SpellCheckerParameter} type input.
        """
        super().__init__(fsm,
                         nGram,
                         parameter)
        self.loadContextDictionaries()

    def loadContextDictionaries(self):
        """
        This method also loads context information from a file.
        """
        self.__context_list = {}
        input_file = open(pkg_resources.resource_filename(__name__, 'data/context_list.txt'), "r", encoding="utf8")
        lines = input_file.readlines()
        for line in lines:
            items = line.strip().split("\t")
            if len(items) == 2:
                word = items[0]
                otherWords = items[1].split(" ")
                self.__context_list[word] = otherWords
        input_file.close()

    def candidateList(self, word: Word, sentence: Sentence) -> list:
        """
        Uses context information to generate candidates for a misspelled word.
        The candidates are the words that are in the context of the neighbouring words of the misspelled word.
        Uses the {@Link damerauLevenshteinDistance(String, String) method to calculate the distance between the misspelled word and
        the candidates and to determine whether the candidates are valid.
        :param word: the misspelled word
        :param sentence: the sentence containing the misspelled word
        :return: an ArrayList of valid candidates for the misspelled word
        """
        words = []
        candidates = set()
        valid_candidates = []
        for w in sentence.getWords():
            if w != word:
                words.append(w)
        for w in words:
            parses: FsmParseList = self.fsm.morphologicalAnalysis(Word.toCapital(w.getName()))
            if parses.size() > 0:
                root = parses.getParseWithLongestRootWord().getWord().getName()
                if self.__context_list.get(root) is not None:
                    for s in self.__context_list[root]:
                        candidates.add(Candidate(s, Operator.CONTEXT_BASED))
        for candidate in candidates:
            if len(candidate.getName()) < 5:
                distance = 1
            elif len(candidate.getName()) < 7:
                distance = 2
            else:
                distance = 3
            if self.damerauLevenshteinDistance(word.getName(), candidate.getName()) <= distance:
                valid_candidates.append(candidate)
        return valid_candidates

    def damerauLevenshteinDistance(self,
                                   first: str,
                                   second: str) -> int:
        """
        Calculates the Damerau-Levenshtein distance between two strings.
        This method also allows for the transposition of adjacent characters,
        which is not possible in a standard Levenshtein distance calculation.
        :param first: the first string
        :param second: the second string
        :return: the Damerau-Levenshtein distance between the two strings
        """
        if first is None or first == "":
            if second is None or second == "":
                return 0
            else:
                return len(second)
        if second is None or second == "":
            return len(first)
        first_length = len(first)
        second_length = len(second)
        distance_matrix = []
        for first_index in range(first_length + 1):
            distance_matrix.append([])
            for second_index in range(second_length + 1):
                distance_matrix[first_index].append(0)
        for first_index in range(first_length + 1):
            distance_matrix[first_index][0] = first_index
        for second_index in range(second_length + 1):
            distance_matrix[0][second_index] = second_index
        for first_index in range(1, first_length + 1):
            for second_index in range(1, second_length + 1):
                if first[first_index - 1] == second[second_index - 1]:
                    cost = 0
                else:
                    cost = 1
                distance_matrix[first_index][second_index] = min(min(distance_matrix[first_index - 1][second_index] + 1,
                                                                  distance_matrix[first_index][second_index - 1] + 1),
                                                              distance_matrix[first_index - 1][second_index - 1] + cost)
                if first_index == 1 or second_index == 1:
                    continue
                if first[first_index - 1] == second[second_index - 2] and first[first_index - 2] == second[
                    second_index - 1]:
                    distance_matrix[first_index][second_index] = min(distance_matrix[first_index][second_index],
                                                                  distance_matrix[first_index - 2][
                                                                      second_index - 2] + cost)
        return distance_matrix[first_length][second_length]



from SpellChecker.Candidate import Candidate
from SpellChecker.Operator import Operator


class TrieCandidate(Candidate):

    __current_index: int
    __current_penalty: float

    def __init__(self,
                 word: str,
                 current_index: int,
                 current_penalty: float):
        """
        Constructs a TrieCandidate object.
        :param word: the candidate word
        :param current_index: the current index of the candidate word
        :param current_penalty: the currentPenalty associated with the candidate word
        """
        super().__init__(word, Operator.TRIE_BASED)
        self.__current_index = current_index
        self.__current_penalty = current_penalty

    def getCurrentIndex(self) -> int:
        """
        Returns the current index of the candidate word.
        :return: the current index of the candidate word
        """
        return self.__current_index

    def getCurrentPenalty(self) -> float:
        """
        Returns the currentPenalty value associated with the candidate word.
        :return: the currentPenalty value associated with the candidate word
        """
        return self.__current_penalty

    def nextIndex(self):
        """
        Increments the current index of the candidate word by 1.
        """
        self.__current_index = self.__current_index + 1



from collections import deque

import pkg_resources
from Corpus.Sentence import Sentence
from Dictionary.Word import Word
from MorphologicalAnalysis.FsmMorphologicalAnalyzer import FsmMorphologicalAnalyzer
from NGram.NGram import NGram

from SpellChecker.Candidate import Candidate
from SpellChecker.NGramSpellChecker import NGramSpellChecker
from SpellChecker.SpellCheckerParameter import SpellCheckerParameter
from SpellChecker.Trie import Trie
from SpellChecker.TrieCandidate import TrieCandidate


class TrieBasedSpellChecker(NGramSpellChecker):
    __generated_words: []
    __trie: Trie

    def __init__(self,
                 fsm: FsmMorphologicalAnalyzer,
                 nGram: NGram,
                 parameter: SpellCheckerParameter):
        """
        A constructor of {@link TrieBasedSpellChecker} class which takes a {@link FsmMorphologicalAnalyzer}, an {@link NGram}
        and a {@link SpellCheckerParameter} as inputs. Then, calls its super class {@link NGramSpellChecker} with given inputs.
        :param fsm: {@link FsmMorphologicalAnalyzer} type input.
        :param nGram: {@link NGram} type input.
        :param parameter: {@link SpellCheckerParameter} type input.
        """
        super().__init__(fsm,
                         nGram,
                         parameter)
        self.loadTrieDictionaries()
        self.prepareTrie()

    def loadTrieDictionaries(self):
        """
        This method also loads generated words from a file.
        """
        self.__generated_words = []
        input_file = open(pkg_resources.resource_filename(__name__, 'data/generated_words.txt'), "r", encoding="utf8")
        lines = input_file.readlines()
        for line in lines:
            self.__generated_words.append(line.strip())
        input_file.close()

    def prepareTrie(self):
        """
        Populates a Trie data structure with a set of strings generated by the morphological analyzer.
        """
        self.__trie = Trie()
        for word in self.__generated_words:
            self.__trie.insert(word)

    def candidateList(self, word: Word, sentence: Sentence) -> list:
        """
        Checks if the trie object contains the candidate.
        If it does, it compares the penalty of the candidate in the results with the penalty of the candidate variable.
        If the current penalty of the candidate in the results is less than or equal to the current penalty
        of the candidate variable, it removes the first element of the candidates ArrayDeque.
        If not, it adds the candidate variable to the results ArrayList and removes the first element.
        If the trie object does not contain the candidate, it checks if the current penalty and the current
        index of the candidate variable and if they are not at the limit,
        it moves the index one character to the right and generates new candidates.
        :param word: the word to search for candidates for
        :param sentence: the sentence containing the word
        :return: an ArrayList of Candidate objects representing the potential candidates
        """
        candidates = deque()
        results = []
        candidates.append(TrieCandidate(word.getName(), -1, 0))
        penalty_limit = min(word.charCount() / 2.0, 3.0)
        while len(candidates) != 0:
            candidate = candidates[0]
            candidate_name = candidate.getName()
            if self.__trie.search(candidate_name):
                item_index = self.searchCandidates(results, candidate)
                if item_index != -1 and results[item_index].getCurrentPenalty() <= candidate.getCurrentPenalty():
                    candidates.popleft()
                else:
                    results.append(candidates.popleft())
            else:
                if candidate.getCurrentPenalty() > penalty_limit - 1 or candidate.getCurrentIndex() >= len(
                        candidate_name) - 1:
                    candidates.popleft()
                else:
                    candidate = candidates.popleft()
                    candidate.nextIndex()
                    candidates.append(
                        TrieCandidate(candidate.getName(), candidate.getCurrentIndex(), candidate.getCurrentPenalty()))
                    new_candidates = self.generateTrieCandidates(candidate)
                    for new_candidate in new_candidates:
                        candidates.append(new_candidate)
        return results

    def searchCandidates(self, result: [Candidate], candidate: TrieCandidate) -> int:
        for i in range(len(result)):
            if result[i].getName() == candidate.getName():
                return i
        return -1

    def generateTrieCandidates(self, candidate: TrieCandidate) -> [TrieCandidate]:
        """
        Generates a set of candidates based on a given TrieCandidate.
        The generated candidates are created by applying a set of operations to
        the input TrieCandidate. The possible operations are:
        De-asciification: replacing certain ASCII characters with their non-ASCII counterparts.
        Substitution: replacing a single character with another character.
        Insertion: adding a single character to the string.
        Deletion: removing a single character from the string.
        Transposition: swapping the positions of two adjacent characters in the string.
        :param candidate: the input TrieCandidate
        :return: a set of candidate strings, each contained in a TrieCandidate object
        """
        candidates = []
        current_name = candidate.getName()
        current_index = candidate.getCurrentIndex()
        current_penalty = candidate.getCurrentPenalty()
        current_node = self.__trie.getTrieNode(current_name[:current_index])
        if current_node is None:
            return candidates
        letters = current_node.childrenToString()
        if current_name[current_index] == 'c':
            ch2 = 'ç'
        elif current_name[current_index] == 'g':
            ch2 = 'ğ'
        elif current_name[current_index] == 'i':
            ch2 = 'ı'
        elif current_name[current_index] == 's':
            ch2 = 'ş'
        elif current_name[current_index] == 'o':
            ch2 = 'ö'
        elif current_name[current_index] == 'u':
            ch2 = 'ü'
        else:
            ch2 = ''
        if ch2 != '':
            deasciified_word = current_name[:current_index] + ch2 + current_name[current_index + 1:]
        else:
            deasciified_word = current_name
        if deasciified_word != current_name and self.__trie.startsWith(deasciified_word[:current_index + 1]):
            candidates.append(TrieCandidate(deasciified_word, current_index, current_penalty + 0.2))
        for j in range(len(letters)):
            replaced = current_name[:current_index] + letters[j] + current_name[current_index + 1:]
            candidates.append(TrieCandidate(replaced, current_index, current_penalty + 1))
            added = current_name[:current_index] + letters[j] + current_name[current_index:]
            candidates.append(TrieCandidate(added, current_index, current_penalty + 1))
            added_last = current_name + letters[j]
            if self.__trie.startsWith(added_last):
                candidates.append(TrieCandidate(added_last, current_index, current_penalty + 1))
        deleted = current_name[:current_index] + current_name[current_index + 1:]
        if len(deleted) > 1:
            candidates.append(TrieCandidate(deleted, current_index - 1, current_penalty + 1))
        if current_index < len(current_name) - 1 and current_name[current_index] != current_name[current_index + 1] \
                and self.__trie.startsWith(current_name[:current_index + 2]):
            swapped = current_name[:current_index] + current_name[current_index + 1] + current_name[current_index] + \
                      current_name[current_index + 2:]
            candidates.append(TrieCandidate(swapped, current_index, current_penalty + 1))
        return candidates



from __future__ import annotations


class TrieNode:
    __children: dict
    __isWord: bool

    def __init__(self):
        """
        A constructor of {@link TrieNode} class which constructs a new TrieNode with an empty children HashMap
        """
        self.__children = {}
        self.__isWord = False

    def getChild(self, ch: str) -> TrieNode:
        """
        Returns the child TrieNode with the given character as its value.
        :param ch: The character value of the child TrieNode.
        :return: TrieNode with the given character value.
        """
        return self.__children.get(ch)

    def addChild(self, ch: str, child: TrieNode):
        self.__children[ch] = child

    def childrenToString(self) -> str:
        result = ""
        for ch in self.__children:
            result = result + ch
        return result

    def isWord(self) -> bool:
        """
        Returns whether the current TrieNode represents the end of a word.
        :return: true if the current TrieNode represents the end of a word, false otherwise.
        """
        return self.__isWord

    def setIsWord(self, isWord: bool):
        """
        Sets whether the current TrieNode represents the end of a word.
        :param isWord: true if the current TrieNode represents the end of a word, false otherwise.
        """
        self.__isWord = isWord



from abc import abstractmethod

from Corpus.Sentence import Sentence


class SpellChecker:

    @abstractmethod
    def spellCheck(self, sentence: Sentence) -> Sentence:
        """
        The spellCheck method which takes a Sentence as an input.

        PARAMETERS
        ----------
        sentence : Sentence
            Sentence type input.

        RETURNS
        -------
        Sentence
            Sentence result.
        """
        pass



import re
from random import randrange

import pkg_resources
from Corpus.Sentence import Sentence
from Dictionary.TxtWord import TxtWord
from Dictionary.Word import Word
from Language.TurkishLanguage import TurkishLanguage
from MorphologicalAnalysis.FsmMorphologicalAnalyzer import FsmMorphologicalAnalyzer
from SpellChecker.Candidate import Candidate
from SpellChecker.Operator import Operator

from SpellChecker.SpellChecker import SpellChecker


class SimpleSpellChecker(SpellChecker):
    fsm: FsmMorphologicalAnalyzer
    __merged_words: dict
    __split_words: dict
    __shortcuts: list = ["cc", "cm2", "cm", "gb", "ghz", "gr", "gram", "hz", "inc", "inch", "inç", "kg", "kw", "kva",
                         "litre", "lt", "m2", "m3", "mah", "mb", "metre", "mg", "mhz", "ml", "mm", "mp", "ms",
                         "mt", "mv", "tb", "tl", "va", "volt", "watt", "ah", "hp", "oz", "rpm", "dpi", "ppm", "ohm",
                         "kwh", "kcal", "kbit", "mbit", "gbit", "bit", "byte", "mbps", "gbps", "cm3", "mm2", "mm3",
                         "khz", "ft", "db", "sn"]
    __conditionalShortcuts: list = ["g", "v", "m", "l", "w", "s"]
    __questionSuffixList: list = ["mi", "mı", "mu", "mü", "miyim", "misin", "miyiz", "midir", "miydi", "mıyım", "mısın",
                                  "mıyız", "mıdır", "mıydı", "muyum", "musun", "muyuz", "mudur", "muydu", "müyüm",
                                  "müsün", "müyüz", "müdür", "müydü", "miydim", "miydin", "miydik", "miymiş", "mıydım",
                                  "mıydın", "mıydık", "mıymış", "muydum", "muydun", "muyduk", "muymuş", "müydüm",
                                  "müydün", "müydük", "müymüş", "misiniz", "mısınız", "musunuz", "müsünüz", "miyimdir",
                                  "misindir", "miyizdir", "miydiniz", "miydiler", "miymişim", "miymişiz", "mıyımdır",
                                  "mısındır", "mıyızdır", "mıydınız", "mıydılar", "mıymışım", "mıymışız", "muyumdur",
                                  "musundur", "muyuzdur", "muydunuz", "muydular", "muymuşum", "muymuşuz", "müyümdür",
                                  "müsündür", "müyüzdür", "müydünüz", "müydüler", "müymüşüm", "müymüşüz", "miymişsin",
                                  "miymişler", "mıymışsın", "mıymışlar", "muymuşsun", "muymuşlar", "müymüşsün",
                                  "müymüşler", "misinizdir", "mısınızdır", "musunuzdur", "müsünüzdür"]

    def __init__(self, fsm: FsmMorphologicalAnalyzer):
        """
        A constructor of SimpleSpellChecker class which takes a FsmMorphologicalAnalyzer as an input and
        assigns it to the fsm variable.

        PARAMETERS
        ----------
        fsm : FsmMorphologicalAnalyzer
            FsmMorphologicalAnalyzer type input.
        """
        self.fsm = fsm
        self.__merged_words = {}
        self.__split_words = {}
        self.loadDictionaries()

    def __generateCandidateList(self, word: str) -> list:
        """
        The generateCandidateList method takes a String as an input. Firstly, it creates a String consists of lowercase
        Turkish letters and an list candidates. Then, it loops i times where i ranges from 0 to the length of given
        word. It gets substring from 0 to ith index and concatenates it with substring from i+1 to the last index as a
        new String called deleted. Then, adds this String to the candidates list. Secondly, it loops j times where j
        ranges from 0 to length of lowercase letters String and adds the jth character of this String between substring
        of given word from 0 to ith index and the substring from i+1 to the last index, then adds it to the candidates
        list. Thirdly, it loops j times where j ranges from 0 to length of lowercase letters String and adds the jth
        character of this String between substring of given word from 0 to ith index and the substring from i to the
        last index, then adds it to the candidates list.

        PARAMETERS
        ----------
        word : str
            String input.

        RETURNS
        -------
        list
            List candidates.
        """
        s = TurkishLanguage.LOWERCASE_LETTERS
        candidates = []
        for i in range(len(word)):
            if i < len(word) - 1:
                swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]
                candidates.append(Candidate(swapped, Operator.SPELL_CHECK))
            if word[i] in TurkishLanguage.LETTERS or word[i] in "wqx":
                deleted = word[:i] + word[i + 1:]
                candidates.append(Candidate(deleted, Operator.SPELL_CHECK))
                for j in range(len(s)):
                    replaced = word[:i] + s[j] + word[i + 1:]
                    candidates.append(Candidate(replaced, Operator.SPELL_CHECK))
                for j in range(len(s)):
                    added = word[:i] + s[j] + word[i:]
                    candidates.append(Candidate(added, Operator.SPELL_CHECK))
        return candidates

    def candidateList(self, word: Word, sentence: Sentence) -> list:
        """
        The candidateList method takes a Word as an input and creates a candidates list by calling generateCandidateList
        method with given word. Then, it loop i times where i ranges from 0 to size of candidates list and creates a
        FsmParseList by calling morphologicalAnalysis with each item of candidates list. If the size of FsmParseList is
        0, it then removes the ith item.

        PARAMETERS
        ----------
        word : Word
            Word input.
        sentence: Sentence
            Sentence input.

        RETURNS
        -------
        list
            candidates list.
        """
        candidates = self.__generateCandidateList(word.getName())
        i = 0
        while i < len(candidates):
            fsm_parse_list = self.fsm.morphologicalAnalysis(candidates[i].getName())
            if fsm_parse_list.size() == 0:
                new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())
                if new_candidate != "" and self.fsm.morphologicalAnalysis(new_candidate).size() > 0:
                    candidates[i] = Candidate(new_candidate, Operator.MISSPELLED_REPLACE)
                else:
                    candidates.pop(i)
                    i = i - 1
            i = i + 1
        return candidates

    def spellCheck(self, sentence: Sentence) -> Sentence:
        """
        The spellCheck method takes a Sentence as an input and loops i times where i ranges from 0 to size of words in
        given sentence. Then, it calls morphologicalAnalysis method with each word and assigns it to the FsmParseList,
        if the size of FsmParseList is equal to the 0, it adds current word to the candidateList and assigns it to the
        candidates list. If the size of candidates greater than 0, it generates a random number and selects an item from
        candidates list with this random number and assign it as newWord. If the size of candidates is not greater than
        0, it directly assigns the current word as newWord. At the end, it adds the newWord to the result Sentence.

        PARAMETERS
        ----------
        sentence : Sentence
            Sentence type input.

        RETURNS
        -------
        Sentence
            Sentence result.
        """
        result = Sentence()
        i = 0
        while i < sentence.wordCount():
            word = sentence.getWord(i)
            next_word = None
            previous_word = None
            if i > 0:
                previous_word = sentence.getWord(i - 1)
            if i < sentence.wordCount() - 1:
                next_word = sentence.getWord(i + 1)
            if self.forcedMisspellCheck(word, result) or \
                    self.forcedBackwardMergeCheck(word, result, previous_word) or \
                    self.forcedSuffixMergeCheck(word, result, previous_word):
                i = i + 1
                continue
            if self.forcedForwardMergeCheck(word, result, next_word) or \
                    self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
                i = i + 2
                continue
            if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \
                    self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
                i = i + 1
                continue
            fsm_parse_list = self.fsm.morphologicalAnalysis(word.getName())
            upper_case_fsm_parse_list = self.fsm.morphologicalAnalysis(Word.toCapital(word.getName()))
            if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:
                candidates = self.mergedCandidatesList(previous_word, word, next_word)
                if len(candidates) < 1:
                    candidates.extend(self.candidateList(word, sentence))
                if len(candidates) < 1:
                    candidates.extend(self.splitCandidatesList(word))
                if len(candidates) > 0:
                    random_candidate = randrange(len(candidates))
                    new_word = Word(candidates[random_candidate])
                    if candidates[random_candidate].getOperator() == Operator.BACKWARD_MERGE:
                        result.replaceWord(i - 1, new_word)
                        i = i + 1
                        continue
                    if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:
                        i = i + 1
                    if candidates[random_candidate].getOperator() == Operator.SPLIT:
                        self.addSplitWords(candidates[random_candidate].getName(), result)
                        i = i + 1
                        continue
                else:
                    new_word = word
            else:
                new_word = word
            result.addWord(new_word)
            i = i + 1
        return result

    def forcedMisspellCheck(self,
                            word: Word,
                            result: Sentence) -> bool:
        forced_replacement = self.fsm.getDictionary().getCorrectForm(word.getName())
        if forced_replacement != "":
            result.addWord(Word(forced_replacement))
            return True
        return False

    def forcedBackwardMergeCheck(self,
                                 word: Word,
                                 result: Sentence,
                                 previousWord: Word) -> bool:
        if previousWord is not None:
            forced_replacement = self.getCorrectForm(
                result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),
                self.__merged_words)
            if forced_replacement != "":
                result.replaceWord(result.wordCount() - 1, Word(forced_replacement))
                return True
        return False

    def forcedForwardMergeCheck(self,
                                word: Word,
                                result: Sentence,
                                nextWord: Word) -> bool:
        if nextWord is not None:
            forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),
                                                     self.__merged_words)
            if forced_replacement != "":
                result.addWord(Word(forced_replacement))
                return True
        return False

    def addSplitWords(self,
                      multiWord: str,
                      result: Sentence):
        words = multiWord.split(" ")
        result.addWord(Word(words[0]))
        result.addWord(Word(words[1]))

    def forcedSplitCheck(self,
                         word: Word,
                         result: Sentence) -> bool:
        forced_replacement = self.getCorrectForm(word.getName(), self.__split_words)
        if forced_replacement != "":
            self.addSplitWords(forced_replacement, result)
            return True
        return False

    def forcedShortcutCheck(self, word: Word, result: Sentence) -> bool:
        shortcut_regex = "(([1-9][0-9]*)|[0])(([.]|[,])[0-9]*)?(" + self.__shortcuts[0]
        for i in range(1, len(self.__shortcuts)):
            shortcut_regex = shortcut_regex + "|" + self.__shortcuts[i]
        shortcut_regex = shortcut_regex + ")"
        conditional_shortcut_regex = "(([1-9][0-9]{0,2})|[0])(([.]|[,])[0-9]*)?(" + self.__conditionalShortcuts[0]
        for i in range(1, len(self.__conditionalShortcuts)):
            conditional_shortcut_regex = conditional_shortcut_regex + "|" + self.__conditionalShortcuts[i]
        conditional_shortcut_regex = conditional_shortcut_regex + ")"
        compiled_expression1 = re.compile(shortcut_regex)
        compiled_expression2 = re.compile(conditional_shortcut_regex)
        if compiled_expression1.fullmatch(word.getName()) or compiled_expression2.fullmatch(word.getName()):
            pair = self.getSplitPair(word)
            forced_replacement = pair[0] + " " + pair[1]
            result.addWord(Word(forced_replacement))
            return True
        return False

    def forcedDeDaSplitCheck(self,
                             word: Word,
                             result: Sentence) -> bool:
        word_name = word.getName()
        capitalized_word_name = Word.toCapital(word_name)
        txt_word = None
        if word_name.endswith("da") or word_name.endswith("de"):
            if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \
                    self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
                new_word_name = word_name[0:len(word_name) - 2]
                fsm_parse_list = self.fsm.morphologicalAnalysis(new_word_name)
                txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))
                if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():
                    if self.fsm.morphologicalAnalysis(new_word_name + "'" + "da").size() > 0:
                        result.addWord(Word(new_word_name + "'" + "da"))
                    else:
                        result.addWord(Word(new_word_name + "'" + "de"))
                    return True
                if fsm_parse_list.size() > 0:
                    txt_word = self.fsm.getDictionary().getWord(
                        fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
                if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():
                    result.addWord(Word(new_word_name))
                    if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
                        if txt_word.notObeysVowelHarmonyDuringAgglutination():
                            result.addWord(Word("de"))
                        else:
                            result.addWord(Word("da"))
                    else:
                        if txt_word.notObeysVowelHarmonyDuringAgglutination():
                            result.addWord(Word("da"))
                        else:
                            result.addWord(Word("de"))
                    return True
        return False

    def forcedSuffixMergeCheck(self,
                               word: Word,
                               result: Sentence,
                               previousWord: Word) -> bool:
        li_list = ["li", "lı", "lu", "lü"]
        lik_list = ["lik", "lık", "luk", "lük"]
        compiled_expression = re.compile("[0-9]+")
        if word.getName() in li_list or word.getName() in lik_list:
            if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
                for suffix in li_list:
                    if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
                            previousWord.getName() + "'" + suffix).size() > 0:
                        result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
                        return True
                for suffix in lik_list:
                    if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(
                            previousWord.getName() + "'" + suffix).size() > 0:
                        result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
                        return True
        return False

    def forcedHyphenMergeCheck(self,
                               word: Word,
                               result: Sentence,
                               previousWord: Word,
                               nextWord: Word) -> bool:
        if word.getName() == "-" or word.getName() == "–" or word.getName() == "—":
            compiled_expression = re.compile("[a-zA-ZçöğüşıÇÖĞÜŞİ]+")
            if previousWord is not None and nextWord is not None and \
                    compiled_expression.fullmatch(previousWord.getName()) and \
                    compiled_expression.fullmatch(nextWord.getName()):
                new_word_name = previousWord.getName() + "-" + nextWord.getName()
                if self.fsm.morphologicalAnalysis(new_word_name).size() > 0:
                    result.replaceWord(result.wordCount() - 1, Word(new_word_name))
                    return True
        return False

    def forcedQuestionSuffixSplitCheck(self,
                                       word: Word,
                                       result: Sentence) -> bool:
        word_name = word.getName()
        if self.fsm.morphologicalAnalysis(word_name).size() > 0:
            return False
        for question_suffix in self.__questionSuffixList:
            if word_name.endswith(question_suffix):
                new_word_name = word_name[0:word_name.rindex(question_suffix)]
                txt_word = self.fsm.getDictionary().getWord(new_word_name)
                if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \
                        and isinstance(txt_word, TxtWord) and not txt_word.isCode():
                    result.addWord(Word(new_word_name))
                    result.addWord(Word(question_suffix))
                    return True
        return False

    def loadDictionaries(self):
        input_file = open(pkg_resources.resource_filename(__name__, 'data/merged.txt'), "r", encoding="utf8")
        lines = input_file.readlines()
        for line in lines:
            items = line.strip().split(" ")
            self.__merged_words[items[0] + " " + items[1]] = items[2]
        input_file.close()
        input_file = open(pkg_resources.resource_filename(__name__, 'data/split.txt'), "r", encoding="utf8")
        lines = input_file.readlines()
        for line in lines:
            items = line.strip().split(" ")
            self.__split_words[items[0]] = items[1] + " " + items[2]
        input_file.close()

    def getCorrectForm(self, wordName: str, dictionary: dict) -> str:
        if wordName in dictionary:
            return dictionary[wordName]
        return ""

    def mergedCandidatesList(self, previousWord: Word, word: Word, nextWord: Word) -> list:
        merged_candidates = []
        backward_merge_candidate = None
        if previousWord is not None:
            backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)
            fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())
            if fsm_parse_list.size() != 0:
                merged_candidates.append(backward_merge_candidate)
        if nextWord is not None:
            forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)
            if backward_merge_candidate is None or \
                    backward_merge_candidate.getName() != forward_merge_candidate.getName():
                fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())
                if fsm_parse_list.size() != 0:
                    merged_candidates.append(forward_merge_candidate)
        return merged_candidates

    def splitCandidatesList(self, word: Word) -> list:
        split_candidates = []
        for i in range(4, len(word.getName()) - 3):
            first_part = word.getName()[0:i]
            second_part = word.getName()[i:]
            fsm_parse_list_first = self.fsm.morphologicalAnalysis(first_part)
            fsm_parse_list_second = self.fsm.morphologicalAnalysis(second_part)
            if fsm_parse_list_first.size() > 0 and fsm_parse_list_second.size() > 0:
                split_candidates.append(Candidate(first_part + " " + second_part, Operator.SPLIT))
        return split_candidates

    def getSplitPair(self, word: Word) -> tuple:
        key = ""
        j = 0
        while j < len(word.getName()):
            if "0" <= word.getName()[j] <= "9":
                key = key + word.getName()[j]
            else:
                break
            j = j + 1
        value = word.getName()[j:]
        return key, value



from Dictionary.Word import Word
from SpellChecker.Operator import Operator


class Candidate(Word):

    __operator: Operator

    def __init__(self,
                 candidate: str,
                 operator: Operator):
        super().__init__(candidate)
        self.__operator = operator

    def getOperator(self) -> Operator:
        return self.__operator



from enum import Enum, auto


class Operator(Enum):

    NO_CHANGE = auto()
    MISSPELLED_REPLACE = auto()
    SPELL_CHECK = auto()
    SPLIT = auto()
    FORWARD_MERGE = auto()
    BACKWARD_MERGE = auto()
    CONTEXT_BASED = auto()
    TRIE_BASED = auto()
